<?php

/**
 * Copyright © 2016 Filoblu S.r.l. All rights reserved.
 */

namespace FiloBlu\Esb\Model\Rest;

use Exception;
use FiloBlu\Esb\Api\Data\AcknowledgeInterface;
use FiloBlu\Esb\Api\ExternalCreditmemoInterface;
use FiloBlu\Esb\Helper\LoggerProvider;
use FiloBlu\OrderAdditionalData\Model\AdditionalDataFactory;
use Magento\Framework\Api\SearchCriteriaBuilder;
use Magento\Framework\App\Config\ScopeConfigInterface;
use Magento\Framework\App\Filesystem\DirectoryList;
use Magento\Framework\Exception\FileSystemException;
use Magento\Framework\Exception\LocalizedException;
use Magento\Framework\Filesystem;
use Magento\Framework\Filesystem\File\Write;
use Magento\Framework\Filesystem\File\WriteInterface;
use Magento\Framework\Phrase;
use Magento\Sales\Api\Data\CreditmemoInterface;
use Magento\Sales\Api\Data\OrderExtensionFactory;
use Magento\Sales\Exception\CouldNotRefundException;
use Magento\Sales\Model\Order\CreditmemoFactory;
use Magento\Sales\Model\OrderRepository;
use Magento\Sales\Model\Service\CreditmemoService;
use Monolog\Logger;

if (!defined('DS')):
    define('DS', DIRECTORY_SEPARATOR);
endif;

/**
 * REST Api to save external creditmemos
 */
class ExternalCreditmemo implements ExternalCreditmemoInterface
{

    /**
     * @var OrderExtensionFactory
     */
    protected $_orderRepository;

    /**
     * @var SearchCriteriaBuilder
     */
    protected $_searchCriteriaBuilder;

    /**
     * @var ScopeConfigInterface
     */
    protected $_scopeConfig;

    /**
     * @var OrderExtensionFactory
     */
    protected $_orderExtensionFactory;

    /**
     * @var CreditmemoFactory
     */
    protected $_creditmemoFactory;

    /**
     * @var CreditmemoService
     */
    protected $_creditmemoService;

    /**
     * @var Filesystem
     */
    protected $_filesystem;

    /**
     * @var Logger
     */
    protected $_logger;

    /**
     * @var AcknowledgeInterface
     */
    protected $_response;

    /**
     * @var AdditionalDataFactory
     */
    protected $_additionalDataFactory;

    /**
     * Init plugin
     *
     * @param ScopeConfigInterface $scopeConfig
     * @param OrderRepository $orderRepository
     * @param OrderExtensionFactory $orderExtensionFactory
     * @param CreditmemoFactory $creditmemoFactory
     * @param CreditmemoService $creditmemoService
     * @param Filesystem $filesystem
     * @param SearchCriteriaBuilder $searchCriteriaBuilder
     * @param LoggerProvider $loggerProvider
     * @param AcknowledgeInterface $response
     * @param AdditionalDataFactory $additionalDataFactory
     */
    public function __construct(
            ScopeConfigInterface $scopeConfig,
            OrderRepository $orderRepository,
            OrderExtensionFactory $orderExtensionFactory,
            CreditmemoFactory $creditmemoFactory,
            CreditmemoService $creditmemoService,
            Filesystem $filesystem,
            SearchCriteriaBuilder $searchCriteriaBuilder,
            LoggerProvider $loggerProvider,
            AcknowledgeInterface $response,
            AdditionalDataFactory $additionalDataFactory
    ) {
        $this->_scopeConfig = $scopeConfig;
        $this->_orderRepository = $orderRepository;
        $this->_orderExtensionFactory = $orderExtensionFactory;
        $this->_creditmemoFactory = $creditmemoFactory;
        $this->_creditmemoService = $creditmemoService;
        $this->_filesystem = $filesystem;
        $this->_searchCriteriaBuilder = $searchCriteriaBuilder;
        $this->_logger = $loggerProvider->getLogger();
        $this->_response = $response;
        $this->_additionalDataFactory = $additionalDataFactory;
    }

    /**
     * Save External Creditmemo data. Save PDF file and additional data on Sales Order.
     *
     * @param CreditmemoInterface $creditmemo_data
     * @return AcknowledgeInterface
     * @throws CouldNotRefundException
     * @throws FileSystemException
     * @throws LocalizedException
     * @throws Exception
     */
    public function save(CreditmemoInterface $creditmemo_data)
    {
        $creditmemoExtensionAttributes = $creditmemo_data->getExtensionAttributes();

        if (null == $creditmemoExtensionAttributes) {
            $message = __('Missing extension attributes');
            $this->_logger->critical($message->render());
            throw new CouldNotRefundException(new Phrase($message->render()));
        }
        if (!$creditmemoExtensionAttributes->getOrderIncrementId()) {
            $message = __('Missing order increment id');
            $this->_logger->critical($message->render());
            throw new CouldNotRefundException(new Phrase($message->render()));
        }
        if (!$creditmemoExtensionAttributes->getDocumentType()) {
            $message = __('Missing document type');
            $this->_logger->critical($message->render());
            throw new CouldNotRefundException(new Phrase($message->render()));
        }
        if (null == $creditmemo_data->getIncrementId()) {
            $message = __('Creditmemo increment id is missing.');
            $this->_logger->critical($message->render());
            throw new CouldNotRefundException(new Phrase($message->render()));
        }
        $orderIncrementId = $creditmemoExtensionAttributes->getOrderIncrementId();
        $creditmemoIncrementId = $creditmemo_data->getIncrementId();
        $documentType = $creditmemoExtensionAttributes->getDocumentType();
        $filename = 'order_' . $orderIncrementId . '_' . $documentType . '_' . $creditmemoIncrementId . '.pdf';
        $filename = preg_replace('/[^0-9A-Za-z.]+/', '_', $filename);
        $documentFolder = $this->_scopeConfig->getValue(
                'esb/print_documents/document_folder'
        );
        $writer = $this->_filesystem->getDirectoryWrite(DirectoryList::VAR_DIR);
        /** @var WriteInterface|Write $file */
        $file = $writer->openFile($documentFolder . DS . $filename, 'w');
        try {
            $file->lock();
            try {
                $file->write(base64_decode($creditmemoExtensionAttributes->getDocumentPdf()));
            } finally {
                $file->unlock();
            }
        } finally {
            $file->close();
        }

        $refund_online = $creditmemoExtensionAttributes->getRefundOnline();
        $creditmemo = [
            'increment_id' => $creditmemo_data->getIncrementId(),
            'document_type' => $creditmemoExtensionAttributes->getDocumentType(),
            'grand_total' => $creditmemo_data->getGrandTotal(),
            'document_pdf' => $filename,
            'refund_online' => $refund_online,
        ];

        $searchCriteria = $this->_searchCriteriaBuilder
                        ->addFilter('increment_id', $creditmemoExtensionAttributes->getOrderIncrementId(), 'eq')->create();
        $orderList = $this->_orderRepository->getList($searchCriteria);

        if (count($orderList) !== 1) {
            $message = __('Cannot find order %1', $creditmemoExtensionAttributes->getOrderIncrementId());
            $this->_logger->critical($message->render());
            throw new CouldNotRefundException(new Phrase($message->render()));
        }
        try {
            $items = $orderList->getItems();
            $order = array_shift($items);

            if (!$order->canCreditmemo()) {
                $message = __('Cannot create creditmemo for order %1', $creditmemoExtensionAttributes->getOrderIncrementId());
                $this->_logger->critical($message->render());
                throw new CouldNotRefundException(new Phrase($message->render()));
            }
            $qtys = [];
            $creditmemo_items = $creditmemo_data['items'];
            foreach ($order->getItems() as $order_item) {
                foreach (($creditmemo_items ?? []) as $key => $creditmemo_item) {
                    if ($order_item->getTypeId() != "configurable" && $order_item->getTypeId() != "bundle" && $creditmemo_item['sku'] === $order_item->getSku()) {
                        $qtys[$order_item->getId()] = $creditmemo_item['qty'];
                        break;
                        //unset($creditmemo_items[$key]);
                    }
                }
            }
            if(empty($qtys)){
                foreach ($order->getItems() as $order_item) {
                    if ($order_item->getTypeId() != "configurable" && $order_item->getTypeId() != "bundle") {
                        $qtys[$order_item->getId()] = 0;
                    }
                }
            }
            $shipping_amount = $creditmemo_data->getShippingInclTax();
            if (abs($shipping_amount - $order->getShippingInclTax()) < 0.5) {
                $shipping_amount = $order->getShippingInclTax();
            }
            $payment_fee_amount = $creditmemoExtensionAttributes->getPaymentFeeAmount();
            $payment_tax_amount = $creditmemoExtensionAttributes->getPaymentTaxAmount();
            if ($payment_tax_amount !== null && $payment_tax_amount > 0) {
                $payment_fee_amount += $payment_tax_amount;
            }
            $order_payment_fee_amount = $order->getMspCodAmount();
            if (abs($payment_fee_amount - $order_payment_fee_amount) < 0.5) {
                $payment_fee_amount = $order_payment_fee_amount;
            }

            $init_data=[
                "qtys" => $qtys,
                "shipping_amount" => $shipping_amount,
                "base_msp_cod_amount" => $payment_fee_amount
            ];

            $adjustment_positive = $creditmemo_data->getBaseAdjustmentPositive();
            if($adjustment_positive){
                $this->_logger->info("Adjustment Positive:".$adjustment_positive);
                $init_data["adjustment_positive"] = $adjustment_positive;
            }

            $adjustment_negative = $creditmemo_data->getBaseAdjustmentNegative();
            if($adjustment_negative){
                $this->_logger->info("Adjustment Negative:".$adjustment_negative);
                $init_data["adjustment_negative"] = $adjustment_negative;
            }

            $creditmemo_obj = $this->_creditmemoFactory->createByOrder(
                    $order,
                    $init_data
            );
            if (abs($creditmemo_obj->getGrandTotal() - $creditmemo_data->getGrandTotal()) > 0.1) {
                $message = __('Grand Total is not matching: %1 <> %2', $creditmemo_obj->getGrandTotal(), $creditmemo_data->getGrandTotal());
                $this->_logger->critical($message->render());
                throw new CouldNotRefundException(new Phrase($message->render()));
            }

            if ($refund_online) {
                $invoices = $order->getInvoiceCollection();
                if ($invoices->getSize() !== 1) {
                    $message = __('No invoices to refund.');
                    $this->_logger->critical($message->render());
                    throw new CouldNotRefundException(new Phrase($message->render()));
                }
                $invoice_obj = $invoices->getFirstItem();
                $creditmemo_obj->setInvoice($invoice_obj);
                $this->_creditmemoService->refund($creditmemo_obj, false);
            } else {
                $this->_creditmemoService->refund($creditmemo_obj, true);
            }

            $additionalDataObject = $this->_additionalDataFactory->create();
            $orderExtensionAttributes = $order->getExtensionAttributes();
            $additionalData = ['external_creditmemos' => [$creditmemo_data->getIncrementId() => $creditmemo]];
            if ($orderExtensionAttributes) {
                if ($orderExtensionAttributes->getFbOad()) {
                    $additionalDataObject = $orderExtensionAttributes->getFbOad();
                    $additionalData = [];
                    foreach ($additionalDataObject->getCustomAttributes() as $customAttribute) {
                        $additionalData[$customAttribute->getAttributeCode()] = $customAttribute->getValue();
                    }
                    if (array_key_exists('external_creditmemos', $additionalData)) {
                        $additionalData['external_creditmemos'][$creditmemo_data->getIncrementId()] = $creditmemo;
                    } else {
                        $additionalData['external_creditmemos'] = [$creditmemo_data->getIncrementId() => $creditmemo];
                    }
                }
            } else {
                $orderExtensionAttributes = $this->_orderExtensionFactory->create();
            }

            foreach ($additionalData as $field => $value) {
                $additionalDataObject->setCustomAttribute($field, $value);
            }

            $orderExtensionAttributes->setFbOad($additionalDataObject);
            $order->setExtensionAttributes($orderExtensionAttributes);
            $order->addStatusHistoryComment(__('External refund has been saved: %1 %2', $documentType, $creditmemoIncrementId), 'closed');
            $order->addStatusHistoryComment(__('Order %1 has been refunded', $order->getIncrementId()), 'closed');
            $this->_orderRepository->save($order);

            $this->_response->setStatus('success');
            $this->_response->setCode($documentType . '_' . $creditmemoIncrementId);
            $this->_response->setMessage(__('Order %1 has been refunded:  %2 %3 has been created', $order->getIncrementId(), $documentType, $creditmemoIncrementId));
            $this->_response->setDescription(__('Magento Creditmemo: %1. Refunded Amount: %2', $creditmemo_obj->getIncrementId(), $creditmemo_obj->getGrandTotal()));
            $this->_response->setFlow('create_refund');
            $this->_response->setOrderIncrementId($order->getIncrementId());
            return $this->_response;
        } catch (Exception $e) {
            $this->_logger->critical($e->getMessage());
            throw $e;
        }
    }
}
